# -*- coding: utf-8 -*-

"""
Figure Plot Entry
=================

"""


# %% IMPORTS
# Built-in imports

# Package imports
from qtpy import QtCore as QC, QtGui as QG, QtWidgets as QW

# GuiPy imports
from guipy import layouts as GL, widgets as GW
from guipy.plugins.figure.widgets.types import PLOT_TYPES
from guipy.widgets import get_modified_signal, set_box_value

# All declaration
__all__ = ['FigurePlotEntry']


# %% CLASS DEFINITIONS
# Create custom class for making a plot entry
# TODO: Allow for individual plots to be toggled (toggled QGroupBox?)
# TODO: Write custom QGroupBox that can have a QComboBox as its title?
class FigurePlotEntry(GW.BaseBox):
    # Signals
    entryNameChanged = QC.Signal(str)
    entryRemoveRequested = QC.Signal()

    # Initialize FigurePlotEntry class
    def __init__(self, index, name, toolbar, parent=None, *args, **kwargs):
        # Save provided index, name and FigureToolbar object
        self.index = index
        self.name = name
        self.toolbar = toolbar

        # Call super constructor
        super().__init__(parent)

        # Set up the plot entry box
        self.init(*args, **kwargs)

    # This function sets up the plot entry
    def init(self):
        # Create layout for this plot entry
        self.create_entry_layout()

    # This function creates the entry layout
    # HINT: Use pd.plot?
    def create_entry_layout(self):
        # Create layout
        layout = GL.QFormLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)
        self.layout = layout

        # Create a name editor layout
        name_layout = GL.QHBoxLayout()
        layout.addRow('Name', name_layout)

        # Create entry name editor
        name_box = GW.QLineEdit()
        name_box.setToolTip("Name of this plot entry")
        set_box_value(name_box, self.name)
        get_modified_signal(name_box).connect(self.entryNameChanged)
        name_layout.addWidget(name_box)
        get_modified_signal(name_box).disconnect(self.modified)
        self.name_box = name_box

        # Add a toolbutton for deleting this plot entry
        del_but = GW.QToolButton()
        del_but.setToolTip("Delete this plot entry")
        get_modified_signal(del_but).connect(self.entryRemoveRequested)
        name_layout.addWidget(del_but)

        # If this theme has a 'remove' icon, use it
        if QG.QIcon.hasThemeIcon('remove'):
            del_but.setIcon(QG.QIcon.fromTheme('remove'))
        # Else, use a standard icon
        else:
            del_but.setIcon(del_but.style().standardIcon(
                QW.QStyle.SP_DialogCloseButton))

        # Create a combobox for choosing a plot type
        plot_types = GW.QComboBox()
        plot_types.addItems(PLOT_TYPES['2D'])
        plot_types.setToolTip("Select the plot type you wish to use for this "
                              "plot entry")
        set_box_value(plot_types, -1)
        get_modified_signal(plot_types).connect(self.set_plot_type)
        layout.addRow('Type', plot_types)
        self.plot_types = plot_types

        # Add a separator
        layout.addSeparator()

        # Create a dummy entry to start off
        self.plot_entry = GW.QWidget()
        layout.addRow(self.plot_entry)

    # This function sets the currently used plot type
    @QC.Slot(str)
    def set_plot_type(self, plot_type):
        # Obtain the index of the current plot entry
        index = self.layout.indexOf(self.plot_entry)

        # Remove this plot entry
        self.layout.removeWidget(self.plot_entry)
        self.plot_entry.close()

        # If the plot_type is empty, create dummy widget
        if not plot_type:
            plot_entry = GW.QWidget()
            entry_name = self.name

        # Else, initialize the requested type
        else:
            # Initialize the proper entry
            plot_type = PLOT_TYPES['2D'][plot_type]
            plot_entry = plot_type(self.toolbar)
            entry_name = "%i_%s" % (self.index, plot_type.PREFIX)

        # Insert the new plot entry
        self.layout.insertRow(index, plot_entry)

        # Save new plot entry as the current entry
        self.plot_entry = plot_entry

        # Save new entry name
        set_box_value(self.name_box, entry_name)

    # Override closeEvent to remove the plot from the figure when closed
    def closeEvent(self, *args, **kwargs):
        # Close the plot_type
        self.plot_entry.close()

        # Call super event
        super().closeEvent(*args, **kwargs)



# -*- coding utf-8 -*-

"""
Widgets
=======
Contains the various different :class:`~PyQt5.QtWidgets.QWidget` subclasses
created for *GuiPy*.

"""


# %% IMPORTS
# Import core modules
from . import base, core
from .base import *
from .core import *

# Import base modules
from . import (
    checkbox, color, combobox, dock, entriesbox, exceptions, genericbox,
    itemsbox, lineedit, radiobutton, spinbox, style, tabbar, text, utils)
from .checkbox import *
from .color import *
from .combobox import *
from .dock import *
from .entriesbox import *
from .exceptions import *
from .genericbox import *
from .itemsbox import *
from .lineedit import *
from .radiobutton import *
from .spinbox import *
from .style import *
from .tabbar import *
from .text import *
from .utils import *

# Import misc module
from . import misc
from .misc import *

# All declaration
__all__ = ['base', 'checkbox', 'color', 'combobox', 'core', 'dock',
           'entriesbox', 'exceptions', 'genericbox', 'itemsbox', 'lineedit',
           'misc', 'radiobutton', 'spinbox', 'style', 'tabbar', 'text',
           'utils']
__all__.extend(base.__all__)
__all__.extend(checkbox.__all__)
__all__.extend(color.__all__)
__all__.extend(combobox.__all__)
__all__.extend(core.__all__)
__all__.extend(dock.__all__)
__all__.extend(entriesbox.__all__)
__all__.extend(exceptions.__all__)
__all__.extend(genericbox.__all__)
__all__.extend(itemsbox.__all__)
__all__.extend(lineedit.__all__)
__all__.extend(misc.__all__)
__all__.extend(radiobutton.__all__)
__all__.extend(spinbox.__all__)
__all__.extend(style.__all__)
__all__.extend(tabbar.__all__)
__all__.extend(text.__all__)
__all__.extend(utils.__all__)

# Author declaration
__author__ = "Ellert van der Velden (@1313e)"



# -*- coding: utf-8 -*-

"""
Line Property
=============

"""


# %% IMPORTS
# Package imports
from matplotlib import rcParams

# GuiPy imports
from guipy import widgets as GW
from guipy.plugins.figure.widgets.types.props import BasePlotProp
from guipy.widgets import set_box_value

# All declaration
__all__ = ['LineProp']


# %% CLASS DEFINITIONS
# Define 'Line' plot property
class LineProp(BasePlotProp):
    """
    Provides the definition of the :class:`~LineProp` plot property.

    This property contains boxes for setting the line style, line width and
    line color.

    """

    # Class attributes
    NAME = "Line"
    DISPLAY_NAME = "Line"
    WIDGET_NAMES = [*BasePlotProp.WIDGET_NAMES, 'line_style_box',
                    'line_width_box', 'line_color_box']

    # This function creates and returns a line style box
    def line_style_box(self):
        """
        Creates a widget box for setting the style of this line plot and
        returns it.

        """

        # Make combobox for linestyles
        line_style_box = GW.LineStyleBox()
        line_style_box.setToolTip("Linestyle to be used for this line plot")

        # Set initial value to the default value in MPL
        set_box_value(line_style_box, rcParams['lines.linestyle'])

        # Return name and box
        return('Style', line_style_box)

    # This function creates and returns a line width box
    def line_width_box(self):
        """
        Creates a widget box for setting the width of this line plot and
        returns it.

        """

        # Make a double spinbox for linewidth
        line_width_box = GW.QDoubleSpinBox()
        line_width_box.setToolTip("Width of the plotted line")
        line_width_box.setRange(0, 9999999)
        line_width_box.setSuffix(" pts")

        # Set initial value to the default value in MPL
        set_box_value(line_width_box, rcParams['lines.linewidth'])

        # Return name and box
        return('Width', line_width_box)

    # This function creates a line color box
    def line_color_box(self):
        """
        Creates a widget box for setting the color of this line plot and
        returns it.

        """

        # Make a color box
        line_color_box = GW.ColorBox()
        line_color_box.setToolTip("Color to be used for this line")

        # Connect 'applying' signal
        self.options.applying.connect(line_color_box.set_default_color)

        # Return name and box
        return('Color', line_color_box)



# -*- coding: utf-8 -*-

"""
Tab Bars
========

"""


# %% IMPORTS
# Built-in imports

# Package imports
from qtpy import QtCore as QC, QtGui as QG, QtWidgets as QW

# GuiPy imports
from guipy import widgets as GW
from guipy.widgets import get_box_value

# All declaration
__all__ = ['EditableTabBar', 'TabNameEditor']


# %% CLASS DEFINITIONS
# Custom QTabBar definition that allows for tab names to be edited
class EditableTabBar(GW.QTabBar):
    # Initialize EditableTabBar class
    def __init__(self, parent=None):
        # Call super constructor
        super().__init__(parent)

        # Set up the tab bar
        self.init()

    # This function sets up the editable tab bar
    def init(self):
        # Connecting signals
        self.tabBarDoubleClicked.connect(self.edit_tab_name)

        # Initialize a tab name editor
        self.name_editor = TabNameEditor(self)

    # This function edits the name of a tab
    @QC.Slot(int)
    def edit_tab_name(self, index):
        """
        Edits the name of the tab given by the provided `index`.

        """

        # Open tab name editor
        self.name_editor(index)


# Custom QLineEdit used for renaming a tab
class TabNameEditor(GW.QLineEdit):
    def __init__(self, tabbar_obj):
        # Save provided tabbar_obj
        self.tabbar = tabbar_obj

        # Call super constructor
        super().__init__(tabbar_obj)

        # Set up name editor
        self.init()

    # This function allows a specified tab's name to be edited
    def __call__(self, index):
        # Save this index
        self.index = index

        # Set the focus to this lineedit
        self.setFocus(True)

        # Obtain the size of the tab
        rect = self.tabbar.tabRect(index)

        # Adjust lineedit size to perfectly match underlying tab
        if(index != self.tabbar.count()-1):
            rect.adjust(1, 1, 0, -1)
        else:
            rect.adjust(1, 1, -1, -1)

        # Set size of editor
        self.setFixedSize(rect.size())

        # Move the editor on top of the tab
        self.move(self.tabbar.mapToGlobal(rect.topLeft()))

        # Obtain the name of the tab
        name = self.tabbar.tabText(index)

        # Set the current name in the editor and select it
        self.setText(name)
        self.selectAll()

        # Show the editor
        self.show()

    # This function sets up the name editor
    def init(self):
        # Install event filter to catch events that should close the popup
        self.installEventFilter(self)

        # Set dialog flags
        self.setWindowFlags(
            QC.Qt.Popup |
            QC.Qt.FramelessWindowHint |
            QC.Qt.NoDropShadowWindowHint)

        # Turn frame off
        self.setFrame(False)

        # Set text margins to line up perfectly with tab
        self.setTextMargins(9, 0, 0, 0)

        # Connect signals
        self.editingFinished.connect(self.set_tab_name)

    # Override eventFilter to filter out clicks, ESC and Enter
    def eventFilter(self, widget, event):
        # Check if the event involves anything for which the popup should close
        if (((event.type() == QC.QEvent.MouseButtonPress) and
             not self.geometry().contains(event.globalPos())) or
            ((event.type() == QC.QEvent.KeyPress) and
             event.key() in (QC.Qt.Key_Escape,
                             QC.Qt.Key_Enter,
                             QC.Qt.Key_Return))):
            # Exit the editor
            self.hide()
            self.setFocus(False)
            return(True)

        # Else, process events as normal
        else:
            return(super().eventFilter(widget, event))

    # This function sets the name of a data table tab
    @QC.Slot()
    def set_tab_name(self):
        """
        Sets the name of the tab that was being edited when this
        :obj:`~TabNameEditor` was called.

        """

        # Obtain the current text in the lineedit
        name = get_box_value(self)

        # If name is not empty, set name
        if name:
            # Set the name of the tab indicated with index
            self.tabbar.setTabText(self.index, name)



# -*- coding utf-8 -*-

"""
Utils
=====
Contains various utility functions and non-widget classes for *GuiPy*.

"""


# %% IMPORTS
# Import base modules

# All declaration
__all__ = []

# Author declaration
__author__ = "Ellert van der Velden (@1313e)"



# -*- coding: utf-8 -*-

"""
Spinboxes
=========

"""


# %% IMPORTS
# Built-in imports

# Package imports
from matplotlib import rcParams
from matplotlib.font_manager import font_scalings
from qtpy import QtCore as QC, QtWidgets as QW

# GuiPy imports
from guipy import layouts as GL, widgets as GW
from guipy.widgets import set_box_value

# All declaration
__all__ = ['DualSpinBox', 'FontSizeBox']


# %% CLASS DEFINITIONS
# Make class with two spinboxes
class DualSpinBox(GW.DualBaseBox):
    """
    Defines the :class:`~DualSpinBox` class.

    """

    # Signals
    modified = QC.Signal([], [int, int], [int, float], [float, int],
                         [float, float])

    # Initialize the DualSpinBox class
    def __init__(self, types=(int, int), sep=None, parent=None):
        """
        Initialize an instance of the :class:`~DualSpinBox` class.

        Optional
        --------
        types : tuple of types ({int; float}). Default: (int, int)
            A tuple containing the type of each spinbox.
        sep : str or None. Default: None
            The string that must be used as a separator between the two
            spinboxes. If *None*, no separator is used.
        parent : :obj:`~PyQt5.QtWidgets.QWidget` object or None. Default: None
            The parent widget to use for this dual spinbox or *None* for no
            parent.

        """

        # Call super constructor
        super().__init__(parent)

        # Create the dual spinbox
        self.init(types, sep)

    # This property returns the default 'modified' signal
    @property
    def default_modified_signal(self):
        return(self.modified.__getitem__(self.types))

    # This function creates the dual spinbox
    def init(self, types, sep):
        """
        Sets up the dual spinbox after it has been initialized.

        """

        # Make dict with different line-edits
        box_types = {
            float: GW.QDoubleSpinBox,
            int: GW.QSpinBox}

        # Save provided types
        self.types = types

        # Create the box_layout
        box_layout = GL.QHBoxLayout(self)
        box_layout.setContentsMargins(0, 0, 0, 0)

        # Create two spinboxes with the provided types
        # LEFT
        left_box = box_types[types[0]]()
        box_layout.addWidget(left_box)
        self.left_box = left_box

        # RIGHT
        right_box = box_types[types[1]]()
        box_layout.addWidget(right_box)
        self.right_box = right_box

        # If sep is not None, create label and add it
        if sep is not None:
            sep_label = GW.QLabel(sep)
            sep_label.setSizePolicy(QW.QSizePolicy.Fixed, QW.QSizePolicy.Fixed)
            box_layout.insertWidget(1, sep_label)

    # This function is automatically called whenever 'modified' is emitted
    @QC.Slot()
    def modified_signal_slot(self):
        # Emit modified signal with proper types
        self.modified[self.types[0], self.types[1]].emit(
            *DualSpinBox.get_box_value(self))


# Make class for setting the font size of a label in a figure
class FontSizeBox(GW.QDoubleSpinBox):
    def __init__(self, *args, **kwargs):
        # Call super constructor
        super().__init__(*args, **kwargs)

        # Set up fontsize box
        self.init()

    # This function sets up the fontsize box after it was initialized
    def init(self):
        # Set some properties
        self.setDecimals(1)
        self.setRange(0, 999)
        self.setSuffix(" pts")

    # This function set the value of this special box
    def set_box_value(self, value, *value_sig):
        # If value is a string, it is a font scaling keyword
        if isinstance(value, str):
            # Obtain actual float fontsize
            value = rcParams['font.size']*font_scalings[value]

        # Set value
        set_box_value(self, value, *value_sig, no_custom=True)



# -*- coding: utf-8 -*-

"""
Data Table Model
================

"""


# %% IMPORTS
# Built-in imports
from itertools import chain
import string

# Package imports
import numpy as np
import pandas as pd
from qtpy import QtCore as QC, QtWidgets as QW

# GuiPy imports

# All declaration
__all__ = ['DataTableModel']


# %% GLOBALS
base_26 = list(string.ascii_uppercase)


# %% CLASS DEFINITIONS
# Define model for the DataTable widget
# HINT: https://doc.qt.io/qt-5/model-view-programming.html
# TODO: Implement fetchMore system?
# TODO: Implement drag/drop system (HINT+#using-drag-and-drop-with-item-views)
# OPTIMIZE: Should Vaex be used for this instead of Pandas?
class DataTableModel(QC.QAbstractTableModel):
    # Signals
    firstColumnInserted = QC.Signal()
    lastColumnRemoved = QC.Signal()
    rowCountChanged = QC.Signal(int)
    columnCountChanged = QC.Signal(int)
    columnNameChanged = QC.Signal(int, str)

    # Initialize DataTableModel class
    def __init__(self, parent=None, *args, **kwargs):
        # Call super constructor
        super().__init__(parent)

        # Set up model
        self.init(*args, **kwargs)

    # Implement delete function
    @QC.Slot()
    def delete(self):
        # Delete all columns in the column list
        self.removeColumns(count=self.columnCount())

    # This function sets up the data table model
    def init(self, import_func=None):
        # Connect signals
        self.destroyed.connect(self.delete)
        self.columnsInserted.connect(self.emitColumnsInsertedSignals)
        self.columnsRemoved.connect(self.emitColumnsRemovedSignals)
        self.rowsInserted.connect(self.emitRowsInsertedSignals)
        self.rowsRemoved.connect(self.emitRowsRemovedSignals)

        # Make a look-up dict for dtypes
        self.dtypes = {
            np.bool_: 'bool',
            np.float64: 'float',
            np.int64: 'int',
            np.object_: 'str'}

        # If import_func is None, initialize an empty table
        if import_func is None:
            # Initialize an empty table
            self._data = pd.DataFrame([])

            # Initialize this data table with a 5x5 table
            self.insertColumns(count=5)
            self.insertRows(count=5)

        # If import_func is not None, call it to initialize the table
        else:
            # Call the function to obtain the data
            self._data = import_func(self)

            # Check if the data frame has the proper column names
            if self._data.columns.dtype.type is np.int64:
                renames = {i: to_base_26(i+1) for i in self._data.columns}
                self._data.rename(columns=renames, inplace=True)

            # Notify other functions that columns have been inserted
            self.beginInsertColumns(QC.QModelIndex(), 0, self.columnCount()-1)
            self.endInsertColumns()

            # Notify other functions that rows have been inserted
            self.beginInsertRows(QC.QModelIndex(), 0, self.rowCount()-1)
            self.endInsertRows()

    # This function emits proper signals when columns have been inserted
    @QC.Slot(QC.QModelIndex, int, int)
    def emitColumnsInsertedSignals(self, parent, first, last):
        # Emit columnCountChanged signal
        self.columnCountChanged.emit(self.columnCount())

        # If first+last+1 is equal to count, emit firstColumnInserted signal
        if(first+last+1 == self.columnCount()):
            self.firstColumnInserted.emit()

    # This function emits proper signals when columns have been removed
    @QC.Slot(QC.QModelIndex, int, int)
    def emitColumnsRemovedSignals(self, parent, first, last):
        # Emit columnCountChanged signal
        self.columnCountChanged.emit(self.columnCount())

        # If columnCount is equal to zero, emit lastColumnRemoved signal
        if not self.columnCount():
            self.lastColumnRemoved.emit()

    # This function emits proper signals when rows have been inserted
    @QC.Slot(QC.QModelIndex, int, int)
    def emitRowsInsertedSignals(self, parent, first, last):
        # Emit rowCountChanged signal
        self.rowCountChanged.emit(self.rowCount())

    # This function emits proper signals when rows have been removed
    @QC.Slot(QC.QModelIndex, int, int)
    def emitRowsRemovedSignals(self, parent, first, last):
        # Emit rowCountChanged signal
        self.rowCountChanged.emit(self.rowCount())

    # This function returns the data column belonging to a specified name/int
    @QC.Slot(int)
    @QC.Slot(str)
    def dataColumn(self, index):
        """
        Returns the :obj:`~pandas.Series` object that belongs to the column
        with the provided column `index`.

        Parameters
        ----------
        index : int or str
            If int, the index of the column whose data is requested.
            If str, the name of this column.

        Returns
        -------
        data_column : :obj:`~pandas.Series`
            The data column that belongs to the column specified by the
            provided `index`.

        """

        # If index is an int, return the column with that index
        if isinstance(index, int):
            return(self._data.iloc[:, index])

        # Else if index is a str, check what column that is and return it
        elif isinstance(index, str):
            return(self._data.loc[:, index])

    # This function returns a list with all data column names
    @QC.Slot()
    def columnNames(self):
        """
        Returns a list with the names of all :obj:`~pandas.Series` objects
        stored in this model.

        Returns
        -------
        names : list of str
            List with names of all data columns.

        """

        # Return list of data column names
        return(list(self._data.columns))

    # Override headerData function
    def headerData(self, section, orientation, role):
        # If role is not DisplayRole, return empty QVariant
        if(role != QC.Qt.DisplayRole):
            return(QC.QVariant())

        # If the horizontal header information is requested
        if(orientation == QC.Qt.Horizontal):
            # Return the corresponding column name
            return(self._data.columns[section])

        # If the vertical header information is requested
        else:
            # Return the corresponding row name
            return(self._data.index[section])

    # Override data function
    def data(self, index, role):
        # If this index is valid
        if index.isValid() and role in (QC.Qt.DisplayRole, QC.Qt.EditRole):
            # Obtain the requested value
            value = self._data.iat[index.row(), index.column()]

            # Convert value to a Python scalar
            if isinstance(value, np.generic):
                value = value.item()

            # If value is None, always use an empty QVariant
            if pd.isna(value):
                data_point = QC.QVariant()

            # Else, create proper QVariant
            else:
                data_point = QC.QVariant(value)

        # Else, use an empty QVariant
        else:
            data_point = QC.QVariant()

        # Return created data_point
        return(data_point)

    # Override flags function
    def flags(self, index):
        # If this index is valid, this item is editable
        if index.isValid():
            return(QC.Qt.ItemIsEnabled |
                   QC.Qt.ItemIsSelectable |
                   QC.Qt.ItemIsEditable)
        # Else, this item is only enabled and selectable
        else:
            return(QC.Qt.ItemIsEnabled |
                   QC.Qt.ItemIsSelectable)

    # Override setData function
    def setData(self, index, value, role):
        # If this index is valid and the role is editing
        if index.isValid() and (role == QC.Qt.EditRole):
            # Set the value
            self._data.iat[index.row(), index.column()] = value

            # Emit dataChanged signal
            self.dataChanged.emit(index, index, [role])

            # Return that operation finished successfully
            return(True)

        # Else, return that operation did not finish successfully
        else:
            return(False)

    # Override rowCount function
    @QC.Slot()
    @QC.Slot(QC.QModelIndex)
    def rowCount(self, parent=None):
        # Return row count
        if self._data.empty:
            return(0)
        else:
            return(self._data.shape[0])

    # Override columnCount function
    @QC.Slot()
    @QC.Slot(QC.QModelIndex)
    def columnCount(self, parent=None):
        # Return column count
        return(self._data.shape[1])

    # This function inserts rows before given row
    # Vaex: df.concat
    @QC.Slot()
    @QC.Slot(int)
    @QC.Slot(int, int)
    @QC.Slot(int, int, QC.QModelIndex)
    def insertRows(self, row=None, count=1, parent=None):
        # If row is None, set it to the current number of rows
        if row is None:
            row = self.rowCount()

        # If parent is None, set it to QC.QModelIndex()
        if parent is None:
            parent = QC.QModelIndex()

        # Notify other functions that rows are going to be inserted
        self.beginInsertRows(parent, row, row+count-1)

        # Create dataframe with the required shape
        insert_df = pd.DataFrame(np.full((count, self.columnCount()), np.nan),
                                 columns=self._data.columns)

        # Concatenate the current dataframe and insert_df
        self._data = pd.concat([self._data[:row], insert_df, self._data[row:]],
                               ignore_index=True)

        # Notify other functions that rows have been inserted
        self.endInsertRows()

        # Emit rowCountChanged signal
        self.rowCountChanged.emit(self.rowCount())

        # Return that operation was successful
        return(True)

    # This function removes rows starting at given row
    # Vaex: df.take + df.to_copy?
    @QC.Slot()
    @QC.Slot(int)
    @QC.Slot(int, int)
    @QC.Slot(int, int, QC.QModelIndex)
    def removeRows(self, row=None, count=1, parent=None):
        # If row is None, set it to the current number of rows-count
        if row is None:
            row = self.rowCount()-count

        # If parent is None, set it to QC.QModelIndex()
        if parent is None:
            parent = QC.QModelIndex()

        # Notify other functions that rows are going to be removed
        self.beginRemoveRows(parent, row, row+count-1)

        # Remove the rows
        indexes = chain(range(0, row), range(row+count, self.rowCount()))
        self._data = self._data.reindex(index=indexes)
        self._data.reset_index(drop=True, inplace=True)

        # Notify other functions that rows have been removed
        self.endRemoveRows()

        # Emit rowCountChanged signal
        self.rowCountChanged.emit(self.rowCount())

        # Return that operation was successful
        return(True)

    # This function clears rows starting at given row
    @QC.Slot(int)
    @QC.Slot(int, int)
    @QC.Slot(int, int, QC.QModelIndex)
    def clearRows(self, row, count=1, parent=None):
        # Clear the rows
        self._data.iloc[row:row+count] = np.nan

        # Return that operation was successful
        return(True)

    # This function inserts columns before given col
    # Vaex: df.add_column
    @QC.Slot()
    @QC.Slot(int)
    @QC.Slot(int, int)
    @QC.Slot(int, int, QC.QModelIndex)
    def insertColumns(self, col=None, count=1, parent=None):
        # If col is None, set it to current number of columns
        if col is None:
            col = self.columnCount()

        # If parent is None, set it to QC.QModelIndex()
        if parent is None:
            parent = QC.QModelIndex()

        # Notify other functions that columns are going to be inserted
        self.beginInsertColumns(parent, col, col+count-1)

        # Create dict with all renames required
        renames = {to_base_26(i+1): to_base_26(i+1+count)
                   for i in range(col, self.columnCount())}

        # Rename all columns that require renaming
        self._data.rename(columns=renames, inplace=True)

        # Create as many columns as required
        for i in reversed(range(col, col+count)):
            self._data.insert(col, to_base_26(i+1), np.nan)

        # Notify other functions that columns have been inserted
        self.endInsertColumns()

        # Return that operation was successful
        return(True)

    # This function removes columns starting at given col
    # Vaex: df.drop
    @QC.Slot()
    @QC.Slot(int)
    @QC.Slot(int, int)
    @QC.Slot(int, int, QC.QModelIndex)
    def removeColumns(self, col=None, count=1, parent=None):
        # If col is None, set it to current number of columns-count
        if col is None:
            col = self.columnCount()-count

        # If parent is None, set it to QC.QModelIndex()
        if parent is None:
            parent = QC.QModelIndex()

        # If count is equal to columnCount, remove all rows first
        if(self.columnCount() == count):
            self.removeRows(count=self.rowCount())

        # Notify other functions that columns are going to be removed
        self.beginRemoveColumns(parent, col, col+count-1)

        # Create dict with all renames required
        renames = {to_base_26(i+1): to_base_26(i+1-count)
                   for i in range(col+count, self.columnCount())}

        # Delete as many columns as required
        for name in self._data.columns[col:col+count]:
            self._data.pop(name)

        # Rename the remaining columns
        self._data.rename(columns=renames, inplace=True)

        # Notify other functions that columns have been removed
        self.endRemoveColumns()

        # Return that operation was successful
        return(True)

    # This function clears columns starting at given col
    @QC.Slot(int)
    @QC.Slot(int, int)
    @QC.Slot(int, int, QC.QModelIndex)
    def clearColumns(self, col, count=1, parent=None):
        # Clear the columns
        self._data.iloc[:, col:col+count] = np.nan

        # Return that operation was successful
        return(True)

    # This function sets the name of a column
    # Vaex: df.rename_column
    @QC.Slot(int, str)
    def setColumnName(self, col, name):
        # If no name was given, use the base name
        if not name:
            name = to_base_26(col+1)

        # Set column name
        self._data.rename(
            columns={self._data.columns[col]: name}, inplace=True)

        # Emit a signal stating that a column changed its name
        self.columnNameChanged.emit(col, name)

    # This function sets the dtype of a column
    # TODO: If auto-conversion is not possible, ask user if the column should
    # be cleared first instead
    @QC.Slot(int, str)
    def setColumnDataType(self, col, dtype):
        # Set the requested data type
        self._data = self._data.astype({self._data.columns[col]: dtype},
                                       copy=False)


# %% FUNCTION DEFINITIONS
# This function converts a value to base-26 using the alphabetical letters
def to_base_26(value):
    """
    Converts a given positive `value` to a base-26 integer, which is made
    up by the letters in the alphabet, and returns it.

    In base-26, the values are defined in the following way (note that
    there is no zero):

    - [1, 26] -> ['A', 'Z'];
    - [27, 52] -> ['AA', 'AZ'];
    - [53, 78] -> ['BA', 'BZ'];
    - [703, 728] -> ['AAA', 'AAZ'], etc.

    Parameters
    ----------
    value : int
        The positive integer that must be converted to base-26.

    Returns
    -------
    result : str
        The provided `value` converted to base-26.

    """

    # Initialize empty result
    result = ''

    # While value is not zero yet
    while value:
        # Decrease by 1 to map [1, 26] to [0, 25]
        value -= 1

        # Determine the next digit of the base-26 value
        result = base_26[(value % 26)] + result

        # True divide by 26 to determine the next digit
        value //= 26

    # Return result
    return(result)



# -*- coding: utf-8 -*-

"""
Config Core
===========
Provides a collection of core definitions that are required for all
configurations in *GuiPy*.

"""


# %% IMPORTS
# Package imports
from sortedcontainers import SortedDict as sdict

# All declaration
__all__ = ['FILE_EXTS', 'FILE_FILTERS', 'FILE_FORMATS', 'FILE_TYPES',
           'register_file_format']


# %% GLOBALS
# Define GuiPy configuration settings dict
CONFIG = sdict()

# Define dicts of file extensions, filters, formats and types
FILE_EXTS = sdict()
FILE_FILTERS = sdict()
FILE_FORMATS = sdict()
FILE_TYPES = sdict()


# %% FUNCTION DEFINITIONS
# This function registers a file format
# TODO: Should the MIME type of the file also be given?
def register_file_format(file_type, file_exts):
    """
    Registers a provided file format with type `file_type` and associated
    extensions `file_exts` for use in *GuiPy*.

    All file formats must be registered with this function in order to be used.

    Parameters
    ----------
    file_type : str
        The type of file that is being registered.
        This is usually the extended acronym of the file extension, like
        `'Portable Document Format'` for PDF-files or
        `'Scalable Vector Graphics'` for SVG-files.
    file_exts : list of str
        A list of all the file extensions that are associated with this file
        format. Most of the time, file formats only have a single file
        extension, like `['.pdf']` for PDF-files or `['.py']` for Python
        scripts. However, JPG-files would for example have `['.jpg', '.jpeg']`.
        If `file_exts` contains multiple entries, the first entry is used as
        the preferred file extension for this file format.

    """

    # Convert all provided file_exts to a single string used for filters
    filter_exts = "*%s" % (' *'.join(file_exts))

    # Create a filter format for this file format
    filter_format = "%s (%s)" % (file_type, filter_exts)

    # Loop over all extensions and add to the proper dicts
    for ext in file_exts:
        FILE_EXTS[ext] = filter_exts
        FILE_FILTERS.setdefault(filter_format, ext)
        FILE_FORMATS[ext] = filter_format
        FILE_TYPES[ext] = file_type



# -*- coding utf-8 -*-

"""
Plot Types
==========

"""


# %% IMPORTS
# Import core modules
from . import base
from .base import *
from . import core
from .core import *

# Import base modules

# Import subpackages
from . import props

# All declaration
__all__ = ['base', 'core', 'props']
__all__.extend(base.__all__)
__all__.extend(core.__all__)

# Author declaration
__author__ = "Ellert van der Velden (@1313e)"



# -*- coding: utf-8 -*-

"""
Base Plugins
============
Provides a collection of custom :class:`~PyQt5.QtWidgets.QWidget` base classes
that allow for certain plugins to be standardized.

"""


# %% IMPORTS
# Built-in imports

# Package imports
from qtpy import QtCore as QC, QtWidgets as QW
from sortedcontainers import SortedDict as sdict

# GuiPy imports
from guipy import widgets as GW
from guipy.config import BaseConfigPage

# All declaration
__all__ = ['BasePlugin', 'BasePluginWidget', 'PluginConfigPage']


# %% CLASS DEFINITIONS
# Define base class for making plugins
class BasePlugin(object):
    # Define class attributes
    TITLE = ''
    CONFIG_PAGES = []
    MENU_ACTIONS = {}
    REQ_PLUGINS = []
    STATUS_WIDGETS = []
    TOOLBARS = []
    TOOLBAR_ACTIONS = {}

    # Initialize plugin
    def __init__(self, req_plugins):
        # Save provided req_plugins
        self.req_plugins = req_plugins

        # Add config pages
        self.add_config_pages()

    # This function initializes and adds the config pages for this plugin
    def add_config_pages(self):
        # Initialize empty dict of config pages
        self.config_pages = sdict()

        # Initialize all config pages of this plugin
        for config_page in self.CONFIG_PAGES:
            # Initialize config_page
            self.config_pages[config_page.NAME] = config_page(self)

    # This function retrieves a config value
    def get_option(self, section, option):
        return(self.config_pages[section].get_option(option))


# Define base class for making plugin widgets
class BasePluginWidget(GW.QWidget, BasePlugin):
    # Define class attributes
    LOCATION = QC.Qt.LeftDockWidgetArea


# Define base class for making plugin config pages
class PluginConfigPage(BaseConfigPage):
    # Initialize plugin config page
    def __init__(self, plugin, parent=None):
        # Save provided plugin
        self.plugin = plugin

        # Call super constructor
        super().__init__(f'plugins/{plugin.TITLE}:{self.NAME}', parent)



# -*- coding: utf-8 -*-

"""
Data Table Selection Model
==========================

"""


# %% IMPORTS
# Built-in imports

# Package imports
from qtpy import QtCore as QC, QtWidgets as QW

# GuiPy imports

# All declaration
__all__ = ['DataTableSelectionModel']


# %% CLASS DEFINITIONS
# Define model for the DataTable widget
# HINT: https://doc.qt.io/qt-5/model-view-programming.html
class DataTableSelectionModel(QC.QItemSelectionModel):
    pass



# -*- coding: utf-8 -*-

"""
Styles
======

"""


# %% IMPORTS
# Package imports
from matplotlib import rcParams
from matplotlib.lines import lineMarkers, lineStyles
from qtpy import QtCore as QC, QtGui as QG, QtWidgets as QW

# GuiPy imports
from guipy import widgets as GW
from guipy.widgets import set_box_value

# All declaration
__all__ = ['LineStyleBox', 'MarkerStyleBox']


# %% CLASS DEFINITIONS
# Make class for setting the linestyle
class LineStyleBox(GW.QComboBox):
    def __init__(self, *args, **kwargs):
        # Call super constructor
        super().__init__(*args, **kwargs)

        # Set up linestyle box
        self.init()

    # This function sets up the linestyle box
    def init(self):
        # Obtain list with all supported linestyles if not existing already
        if not hasattr(self, 'LINESTYLES'):
            # Create list of all supported linestyles
            linestyles = [(key, value[6:]) for key, value in lineStyles.items()
                          if value != '_draw_nothing']
            linestyles.append(('', 'nothing'))
            linestyles.sort(key=lambda x: x[0])

            # Save as class attribute
            LineStyleBox.LINESTYLES = linestyles

        # Populate this box with all supported linestyles
        for i, (linestyle, tooltip) in enumerate(self.LINESTYLES):
            self.addItem(linestyle)
            self.setItemData(i, tooltip, QC.Qt.ToolTipRole)

        # Set initial value to the default value in MPL
        set_box_value(self, rcParams['lines.linestyle'])

    # Override set_box_value to account for other spellings of 'nothing'
    def set_box_value(self, value, *value_sig):
        # Set value to empty if it uses a different spelling
        if value.lower() in ('none', ' '):
            value = ''
        elif value.lower() in ('dashed', ):
            value = '--'
        elif value.lower() in ('dotted', ):
            value = ':'

        # Call normal method
        set_box_value(self, value, *value_sig, no_custom=True)


# Make class for setting the markerstyle
class MarkerStyleBox(GW.QComboBox):
    def __init__(self, *args, **kwargs):
        # Call super constructor
        super().__init__(*args, **kwargs)

        # Set up markerstyle box
        self.init()

    # This function sets up the markerstyle box
    def init(self):
        # Obtain list with all supported markerstyles if not existing already
        if not hasattr(self, 'MARKERS'):
            # Create list of all supported markerstyles
            markers = [(key, value) for key, value in lineMarkers.items()
                       if(value != 'nothing' and isinstance(key, str))]
            markers.append(('', 'nothing'))
            markers.sort(key=lambda x: x[0])

            # Save as class attribute
            MarkerStyleBox.MARKERS = markers

        # Populate this box with all supported markerstyles
        for i, (marker, tooltip) in enumerate(self.MARKERS):
            self.addItem(marker)
            self.setItemData(i, tooltip, QC.Qt.ToolTipRole)

        # Set initial value to the default value in MPL
        set_box_value(self, rcParams['lines.marker'])

    # Override set_box_value to account for other spellings of 'nothing'
    def set_box_value(self, value, *value_sig):
        # Set value to empty if it uses a different spelling
        if value.lower() in ('none', ' '):
            value = ''

        # Call normal method
        set_box_value(self, value, *value_sig, no_custom=True)



# -*- coding utf-8 -*-

"""
Globals
=======
Provides a collection of all global variables for *GuiPy* that must be
available.

"""


# %% IMPORTS
# Package imports
import numpy as np
from os import path

# All declaration
__all__ = ['APP_NAME', 'CONFIG_DIR', 'CONFIG_NAME', 'DIR_PATH', 'FLOAT_TYPES',
           'INT_TYPES', 'STR_TYPES']


# %% GUI GLOBALS
APP_NAME = 'GuiPy'                                  # Name of application
CONFIG_DIR = '.guipy'                               # Name of config directory
CONFIG_NAME = 'guipy.ini'                           # Name of config file
DIR_PATH = path.abspath(path.dirname(__file__))     # Path to GUI directory


# %% TYPE GLOBALS
INT_TYPES = (int, np.integer)
FLOAT_TYPES = (float, np.floating, *INT_TYPES)
STR_TYPES = (str,)



# -*- coding: utf-8 -*-

"""
Miscellaneous
=============

"""


# %% IMPORTS
# Built-in imports

# Package imports

# GuiPy imports
from guipy import widgets as GW

# All declaration
__all__ = ['type_box_dict']


# %% INSTANCES
# Make a look-up dict for types
type_box_dict = {
    bool: GW.QCheckBox,
    float: GW.FloatLineEdit,
    int: GW.IntLineEdit,
    str: GW.QLineEdit,
    dict: GW.EditableEntriesBox,
    list: GW.GenericItemsBox}



# -*- coding: utf-8 -*-

"""
Base Config
===========
Provides a collection of base functions to standardize the configuration of
*GuiPy*.

"""


# %% IMPORTS
# Package imports
from qtpy import QtCore as QC, QtGui as QG, QtWidgets as QW
from sortedcontainers import SortedDict as sdict

# GuiPy imports
from guipy import widgets as GW
from guipy.config.manager import CONFIG
from guipy.widgets import get_box_value, get_modified_signal, set_box_value

# All declaration
__all__ = ['BaseConfigPage']


# %% CLASS DEFINITIONS
# Define base class for making config pages
class BaseConfigPage(GW.BaseBox):
    # Define class attributes
    NAME = ''

    # Initialize config page
    def __init__(self, section_name=None, parent=None):
        # Save the config section name
        self.section_name = self.NAME if section_name is None else section_name

        # Initialize empty config_entries dict
        self.config_entries = sdict()

        # Initialize restart_flag
        self.restart_flag = False

        # Call super constructor
        super().__init__(parent)

        # Set up config page
        self.init()

        # Add this config page to the config manager
        CONFIG.add_config_page(self)

    # This function sets up the config page
    def init(self):
        pass

    # This function retrieves a config value
    def get_option(self, option):
        return(CONFIG.get_option(self.section_name, option))

    # This function adds the provided widget as a config entry
    def add_config_entry(self, name, widget, restart_flag=False):
        """
        Adds the provided `widget` as a config entry to this config page with
        given `name`.
        This allows for the values of `widget` to be tracked and potentially
        discarded/reverted.

        """

        # Make sure that name is lowercase
        name = name

        # Add widget as an entry to config_entries
        self.config_entries[name] = widget
        get_modified_signal(widget).connect(self.modified)

        # If restart_flag is True, connect set_restart_flag as well
        if restart_flag:
            get_modified_signal(widget).connect(self.set_restart_flag)

    # This function sets restart_flag to True when called
    def set_restart_flag(self):
        self.restart_flag = True

    # This function parses and processes a config section, and returns it
    def decode_config(self, section_dict):
        """
        Parses a section of the config parser, converting it into the values as
        used by *GuiPy* and returns it.

        Parameters
        ----------
        section_dict : dict
            Dict containing the config section belonging to this config page.

        Returns
        -------
        config_dict : dict
            Dict containing the processed config section values. This dict is
            used in *GuiPy* for getting/setting values.

        """

        raise NotImplementedError(self.__class__)

    # This function returns a dict containing the default config values
    def get_default_config(self):
        """
        Returns the default values for the config section belonging to this
        config page.

        Returns
        -------
        default_dict : dict
            Dict containing the default config section values.

        """

        raise NotImplementedError(self.__class__)

    # This function returns its config section, as required by config parser
    def encode_config(self, config_dict):
        """
        Returns a dict containing the config values for the config section
        associated with this config page.

        Parameters
        ----------
        config_dict : dict
            Dict containing the config values belonging to this config page as
            used by *GuiPy*.

        Returns
        -------
        section_dict : dict
            Dict containing the config section belonging to this config page.
            This dict is used by the config parser to save config values to
            file.

        """

        raise NotImplementedError(self.__class__)

    # This function applies the currently stored config
    def apply_config(self, config_dict):
        """
        Applies the config values for the config section associated with this
        config page.

        This function is called whenever the currently stored config values
        must be applied to the various settings they are associated with.

        Parameters
        ----------
        config_dict : dict
            Dict containing the config values belonging to this config page as
            used by *GuiPy*.

        """

        pass

    # This function returns a dict containing the current config values
    def get_box_value(self, *value_sig):
        """
        Returns the current values for the config section belonging to this
        config page.

        Returns
        -------
        config_dict : dict
            Dict containing the current config section values.

        """

        # Create empty dict of config values
        config_dict = sdict()

        # Retrieve values of all config entries
        for key, entry_box in self.config_entries.items():
            config_dict[key] = get_box_value(entry_box)

        # Return the config_dict
        return(config_dict)

    # This function sets a config dict as the new config values
    def set_box_value(self, value, *value_sig):
        """
        Sets the values in the provided config dict `value` as the new values
        of the config section belonging to this config page.

        Parameters
        ----------
        value : dict
            Dict containing the new config section values.

        """

        # Assign config_dict
        config_dict = value

        # Set all values contained in the config_dict
        for key, value in config_dict.items():
            entry_box = self.config_entries[key]
            set_box_value(entry_box, value)



# -*- coding: utf-8 -*-

"""
Radiobuttons
============

"""


# %% IMPORTS
# Built-in imports
from itertools import repeat

# Package imports
import numpy as np
from qtpy import QtCore as QC, QtWidgets as QW

# GuiPy imports
from guipy import INT_TYPES, STR_TYPES
from guipy import layouts as GL, widgets as GW
from guipy.widgets import get_box_value, set_box_value

# All declaration
__all__ = ['MultiRadioButton']


# %% CLASS DEFINITIONS
# Make class with N RadioButtons
class MultiRadioButton(GW.BaseBox):
    """
    Defines the :class:`~MultiRadioButton` class.

    """

    # Signals
    modified = QC.Signal([], [int], [str])

    # Initialize the MultiRadioButton class
    def __init__(self, n=2, layout='horizontal', parent=None):
        """
        Initialize an instance of the :class:`~MultiRadioButton` class.

        Optional
        --------
        n : int or list of str. Default: 2
            The number of radiobuttons this multi-radiobutton must have.
            If list of str, the list contains the names of all radiobuttons
            that must be created.
        layout : {'horizontal'; 'vertical'} or tuple of (n_rows, n_cols). \
            Default: 'horizontal'
            The layout that this multi-radiobutton must have.
            If 'horizontal' or 'vertical', all radiobuttons will be aligned on
            a single row or column, respectively.
            If tuple, the given `n_rows` and `n_cols` determine the shape of
            the grid. If `n` is greater than `n_rows`*`n_cols`, it will be
            clipped.
        parent : :obj:`~PyQt5.QtWidgets.QWidget` object or None. Default: None
            The parent widget to use for this multi-radiobutton or *None* for
            no parent.

        """

        # Call super constructor
        super().__init__(parent)

        # Create the multi-radiobutton
        self.init(n, layout)

    # Override __getitem__ to return the requested radiobutton(s)
    def __getitem__(self, key):
        # If key is an integer, return the corresponding radiobutton
        if isinstance(key, INT_TYPES):
            # Try to return the requested radiobutton
            try:
                return(self.buttons[key])
            # If that cannot be done, raise IndexError
            except IndexError:
                raise IndexError("Index out of range")

        # If key is a slice object, return everything that is requested
        elif isinstance(key, slice):
            return(*map(self.__getitem__, range(*key.indices(self.N))),)

        # If key is a string, try to return the corresponding radiobutton
        elif isinstance(key, STR_TYPES):
            # Count how many times the requested name appears
            n_hits = self.names.count(key)

            # Check if this name exists
            if n_hits:
                # If so, check if it appears exactly once
                if(n_hits == 1):
                    # If so as well, return the corresponding radiobutton
                    return(self[self.names.index(key)])
                else:
                    # If not, raise error
                    raise KeyError("Key %r is not unique" % (key))
            else:
                # If name does not exist, raise error
                raise KeyError("Key %r not found" % (key))

        # Else, raise TypeError
        else:
            raise TypeError("Index must be of type 'int'; 'slice' or 'str', "
                            "not type %r" % (type(key).__name__))

    # This property returns the default 'modified' signal
    @property
    def default_modified_signal(self):
        return(self.modified[str])

    # This function sets up the multi-radiobutton
    def init(self, n, layout):
        """
        Sets up the multi-radiobutton after it has been initialized.

        """

        # Check what type of n has been provided
        if isinstance(n, INT_TYPES):
            # If n is an integer, set all names to empty strings
            names = ['']*n
        elif hasattr(n, '__iter__') and not isinstance(n, STR_TYPES):
            # Else, n must be an iterable of names
            names = list(map(str, n))
            n = len(n)
        else:
            raise TypeError

        # Check what type of layout has been provided
        if isinstance(layout, STR_TYPES):
            # If layout is a string, it is either 'horizontal' or 'vertical'
            if layout in ('h', 'horizontal', 'r', 'row'):
                layout = GL.QHBoxLayout(self)
            elif layout in ('v', 'vertical', 'c', 'col', 'column'):
                layout = GL.QVBoxLayout(self)
            else:
                raise ValueError
        else:
            # Else, layout is a tuple specifying the number of rows and columns
            n_rows, n_cols = layout
            layout = GL.QGridLayout(self)

        # Set contents margins of layout
        layout.setContentsMargins(0, 0, 0, 0)

        # Initialize empty list of radiobuttons
        self.buttons = []

        # Create buttons iterator
        if isinstance(layout, QW.QGridLayout):
            iterator = zip(names, np.ndindex(n_rows, n_cols))
        else:
            iterator = zip(names, repeat(()))

        # Create all requested radiobuttons
        for item in iterator:
            # Create radiobutton
            button = GW.QRadioButton(item[0])

            # Add button to list and layout
            self.buttons.append(button)
            layout.addWidget(button, *item[1])

        # Save the number of radiobuttons
        self.N = len(self.buttons)

        # Save the names
        self.names = names[:self.N]

    # This function is automatically called whenever 'modified' is emitted
    @QC.Slot()
    def modified_signal_slot(self):
        # Emit modified signal with proper types
        self.modified[int].emit(MultiRadioButton.get_box_value(self, int))
        self.modified[str].emit(MultiRadioButton.get_box_value(self, str))

    # This function retrieves a value of this special box
    def get_box_value(self, *value_sig):
        """
        Returns the index or text of the current radiobutton that is set to
        *True*.

        Returns
        -------
        value : int or str
            The index or text of the current radiobutton that is set to *True*.

        """

        # Obtain the index of the current radiobutton
        index = np.argmax(list(map(get_box_value, self.buttons)))

        # If value_sig is not int, return its text
        if int not in value_sig:
            return(get_box_value(self[index], str))
        # Else, return its index
        else:
            return(index)

    # This function sets the value of this special box
    def set_box_value(self, value, *value_sig):
        """
        Sets the radiobutton with index `value` to *True*.

        Parameters
        ----------
        value : int
            The index of the radiobutton that must be set to *True*.

        """

        set_box_value(self[value], True)



# -*- coding: utf-8 -*-

"""
Plot Types Core
===============
Collects all the registered plot types for figures into a single dict.

"""


# %% IMPORTS
# Built-in imports
from importlib import import_module
import os
from os import path

# Package imports
from sortedcontainers import SortedDict as sdict

# All declaration
__all__ = ['PLOT_TYPES', 'register_plot_type']


# %% GLOBALS
# Define dict of plot types
PLOT_TYPES = {
    '2D': sdict(),
    '3D': sdict()}


# %% FUNCTION DEFINITIONS
# This function registers a plot_type
def register_plot_type(plot_type_class):
    """
    Registers a provided plot type `plot_type_class` for use in *GuiPy*.

    All plot types must be registered with this function in order to be used.

    Parameters
    ----------
    plot_type_class : \
        :class:`~guipy.plugins.figure.widgets.types.BasePlotType` subclass
        The plot type class to use for drawing a given plot.

    """

    # Register the plot_type
    name = plot_type_class.NAME
    axis_type = plot_type_class.AXIS_TYPE
    PLOT_TYPES[axis_type][name] = plot_type_class


# This function imports all pre-defined plot types and registers them
def _import_plot_types():
    """
    Imports and registers all pre-defined plot types for use in *GuiPy*.

    """

    # Obtain the path to this directory
    dirpath = path.dirname(__file__)

    # Obtain a list of all files in this directory
    filenames = next(os.walk(dirpath))[2]

    # Remove __init__.py, base.py and core.py
    filenames.remove('__init__.py')
    filenames.remove('base.py')
    filenames.remove('core.py')

    # Loop over all files remaining
    for filename in filenames:
        # Obtain full module name
        modname = "%s.%s" % (__package__, filename[:-3])

        # Import this module
        mod = import_module(modname)

        # Register everything in __all__ as a plot type
        for type_name in mod.__all__:
            type_class = getattr(mod, type_name)
            register_plot_type(type_class)


# Import all defined plot types
_import_plot_types()



# -*- coding: utf-8 -*-

"""
Data Property
=============

"""


# %% IMPORTS
# Package imports
from qtpy import QtCore as QC, QtGui as QG, QtWidgets as QW

# GuiPy imports
from guipy import widgets as GW
from guipy.plugins.figure.widgets.types.props import BasePlotProp
from guipy.widgets import get_box_value, get_modified_signal, set_box_value

# All declaration
__all__ = ['Data1DProp', 'Data2DProp', 'Data1or2DProp', 'Data3DProp',
           'MultiData1DProp', 'MultiData2DProp', 'MultiData3DProp']


# %% CLASS DEFINITIONS
# Define 'Data1D' plot property
class Data1DProp(BasePlotProp):
    """
    Provides the definition of the :class:`~Data1DProp` plot property.

    This property contains boxes for setting the label and the X-axis data.

    """

    # Class attributes
    NAME = "Data1D"
    DISPLAY_NAME = "Data"
    REQUIREMENTS = [*BasePlotProp.REQUIREMENTS, 'data_table_plugin']
    WIDGET_NAMES = [*BasePlotProp.WIDGET_NAMES, 'data_label_box', 'x_data_box']

    # This function creates and returns the data label box
    def data_label_box(self):
        """
        Creates a widget box for setting the label of the plot data and returns
        it.

        """

        # Make a lineedit for setting the label of the plot
        data_label_box = GW.QLineEdit()
        data_label_box.setToolTip("Label of this plot")

        # Return name and box
        return('Label', data_label_box)

    # This function creates and returns the x-axis data box
    def x_data_box(self):
        """
        Creates a widget box for setting the data for the X-axis and returns
        it.

        """

        # Make a combobox for setting the x-axis data
        x_data_box = DataColumnBox(self.data_table_plugin)
        x_data_box.setToolTip("Data table and column to use for the X-axis "
                              "data")

        # Return name and box
        return('X-axis', x_data_box)


# Define 'Data2D' plot property
class Data2DProp(Data1DProp):
    """
    Provides the definition of the :class:`~Data2DProp` plot property.

    This property contains boxes for setting the label and the X-axis; and
    Y-axis data.

    """

    # Class attributes
    NAME = "Data2D"
    WIDGET_NAMES = [*Data1DProp.WIDGET_NAMES, 'y_data_box']

    # This function creates and returns the y-axis data box
    def y_data_box(self):
        """
        Creates a widget box for setting the data for the Y-axis and returns
        it.

        """

        # Make a combobox for setting the y-axis data
        y_data_box = DataColumnBox(self.data_table_plugin)
        y_data_box.setToolTip("Data table and column to use for the Y-axis "
                              "data")

        # Return name and box
        return('Y-axis', y_data_box)


# Define 'Data1or2D' plot property
class Data1or2DProp(Data2DProp):
    """
    Provides the definition of the :class:`~Data1or2DProp` plot property.

    This property contains boxes for setting the label and the X-axis; and
    Y-axis data. The X-axis data is optional.

    """

    # Class attributes
    NAME = "Data1or2D"

    # This function creates and returns the optional x-axis data box
    def x_data_box(self):
        """
        Creates a widget box for optionally setting the data for the X-axis and
        returns it.

        """

        # Make a combobox for setting the x-axis data
        x_data_box = GW.ToggleBox(
            DataColumnBox(self.data_table_plugin),
            tooltip="Check to manually set the X-axis data")
        x_data_box.setToolTip("Data table and column to use for the X-axis "
                              "data")
        set_box_value(x_data_box, (False, (None, None)))

        # Return name and box
        return('X-axis', x_data_box)


# Define 'Data3D' plot property
class Data3DProp(Data2DProp):
    """
    Provides the definition of the :class:`~Data3DProp` plot property.

    This property contains boxes for setting the label and the X-axis; Y-axis;
    and Z-axis data.

    """

    # Class attributes
    NAME = "Data3D"
    WIDGET_NAMES = [*Data2DProp.WIDGET_NAMES, 'z_data_box']

    # This function creates and returns the z-axis data box
    def z_data_box(self):
        """
        Creates a widget box for setting the data for the Z-axis and returns
        it.

        """

        # Make a combobox for setting the z-axis data
        z_data_box = DataColumnBox(self.data_table_plugin)
        z_data_box.setToolTip("Data table and column to use for the Z-axis "
                              "data")

        # Return name and box
        return('Z-axis', z_data_box)


# Define 'MultiDataND' plot property, which holds several 'DataND' props
class MultiDataNDProp(BasePlotProp):
    """
    Provides the definition of the :class:`~MultiDataNDProp` plot property.

    This property contains a tab widget with multiple `DataNDProp` properties.

    """

    # Class attributes
    DISPLAY_NAME = "Data"
    WIDGET_NAMES = [*BasePlotProp.WIDGET_NAMES, 'multi_data_box']
    TRACK_VALUES = False

    # Initialize multi data property
    def __init__(self, data_prop, *args, **kwargs):
        # Save the provided data_prop
        self.data_prop = data_prop

        # Call super constructor
        super().__init__(*args, **kwargs)

    # This function creates and returns the multi data box
    def multi_data_box(self):
        """
        Creates a widget box for setting the data for multiple 'DataND' props
        and returns it.

        """

        # Make a tab widget for holding data tabs
        tab_widget = MultiDataTabWidget()
        self.tab_widget = tab_widget

        # Create add button for tabs
        add_but = GW.QToolButton()
        add_but.setToolTip("Add additional data set")
        tab_widget.setCornerWidget(add_but, QC.Qt.TopRightCorner)

        # If this theme has an 'add' icon, use it
        if QG.QIcon.hasThemeIcon('add'):
            add_but.setIcon(QG.QIcon.fromTheme('add'))
        # Else, use a simple plus
        else:
            add_but.setText('+')

        # Connect tab widget signals
        get_modified_signal(add_but).connect(self.add_data_box)
        tab_widget.tabCloseRequested.connect(self.remove_data_box)

        # Add initial data box
        self.add_data_box()

        # Return box
        return(tab_widget,)

    # This function adds a data box to the tab widget
    @QC.Slot()
    def add_data_box(self):
        """
        Adds a new data tab to this plot property.

        """

        # Create a default widget for the new DataND prop
        data_prop = GW.BaseBox()
        get_modified_signal(data_prop).connect(self.tab_widget.modified)

        # Create a dictionary with all requirements of this property
        prop_kwargs = {req: getattr(self, req)
                       for req in self.data_prop.REQUIREMENTS}

        # Create the DataND prop
        prop_layout = self.data_prop(**prop_kwargs)

        # Set prop_layout as the layout for data_prop
        data_prop.setLayout(prop_layout)

        # Obtain the name of this data_prop
        name = "data_%i" % (self.tab_widget.count())

        # Add data_prop to the tab widget
        index = self.tab_widget.addTab(data_prop, name)

        # Switch focus to the new tab
        set_box_value(self.tab_widget, index)

        # Check if there is now more than a single tab
        self.tab_widget.setTabsClosable(self.tab_widget.count() > 1)

    # This function removes a data box from the tab widget
    @QC.Slot(int)
    def remove_data_box(self, index):
        """
        Removes the data tab associated with the provided `index` from this
        plot property.

        """

        # Obtain the DataND prop associated with this index
        data_prop = self.tab_widget.widget(index)

        # Close this data_prop
        data_prop.close()

        # Remove this data_prop from the tab widget
        self.tab_widget.removeTab(index)

        # Check if there is still more than a single tab
        self.tab_widget.setTabsClosable(self.tab_widget.count() > 1)


# Define custom QTabWidget for holding the multi data
class MultiDataTabWidget(GW.QTabWidget):
    # Define signals
    modified = QC.Signal()

    # Override constructor to connect some signals
    def __init__(self, *args, **kwargs):
        # Call super constructor
        super().__init__(*args, browse_tabs=True, **kwargs)

        # Connect signals
        self.tabWasInserted.connect(self.modified)
        self.tabWasRemoved.connect(self.modified)

    # Define special get_box_value method
    def get_box_value(self, *value_sig):
        """
        Returns the values stored in the tabs of this multi data tab widget, as
        requested by the provided `value_sig`.

        Parameters
        ----------
        value_sig : positional arguments of object
            The signature of the value(s) that must be returned.
            If a signature is invalid, only the valid part will be used.
            See ``Notes`` for all valid signatures.

        Returns
        -------
        value : obj
            The requested value(s).

        Notes
        -----
        The following signatures are valid (`values` is equal to
        ``[widgets.items() for widgets in tabWidgets]``):

            =================== ===========================================
            value signature     requested value
            =================== ===========================================
            `()`                ``values``
            `(int,)`            ``values[int]``
            `(str,)`            ``[value[str] for value in values]``
            `(int, str)`        ``values[int][str]``
            `(str, *args)`      ``[value[str][*args] for value in values]``
            `(int, str, *args)` ``values[int][str][*args]``
            =================== ===========================================

        """

        # Convert value_sig to a list
        value_sig = list(value_sig)

        # Create empty list of values
        value = []

        # Check if the first argument of value_sig is an integer
        if value_sig and isinstance(value_sig[0], int):
            # If so, use the tab indicated by this index
            index = value_sig.pop(0)
            tabWidgets = [self.widget(index)]
            indexed = True
        else:
            # If not, use all tabs
            tabWidgets = self.tabWidgets()
            indexed = False

        # Loop over all tab widgets
        for widget in tabWidgets:
            # Obtain the dict of widgets in this tab
            widgets = widget.layout().widgets

            # Create dict with all box values of these widgets
            values = {key: get_box_value(box) for key, box in widgets.items()}

            # Add this dict to value
            value.append(values)

        # Check if the first argument of value_sig is a string
        if value_sig and isinstance(value_sig[0], str):
            # If so, it indicates a box that must be returned
            key = value_sig.pop(0)
            value = [val[key] for val in value]

            # If value_sig still has elements, it indexes this box
            for sig in value_sig:
                value = [val[sig] for val in value]

        # If indexed is True, return solely the first argument
        if indexed:
            value = value[0]

        # Return value
        return(value)

    # Define special set_box_value method
    def set_box_value(self, value, *value_sig):
        """
        Sets the provided `value` in the tabs of this multi data tab widget, as
        requested by the provided `value_sig`.

        Parameters
        ----------
        value : obj
            The value that must be set.
        value_sig : positional arguments of object
            The signature of the value(s) that must be set.
            If a signature is invalid, only the valid part will be used.
            See ``Notes`` for all valid signatures.

        Notes
        -----
        The following signatures are valid:

            =============== ========== ===============================
            value signature value type operation
            =============== ========== ===============================
            `()`            -          <default>
            `(int,)`        dict       ``tabWidgets[int][key]``
            `(str,)`        array_like ``tabWidgets[:][str]``
            `(int, str)`    obj        ``tabWidgets[int][str]``
            =============== ========== ===============================

        """

        # If value_sig is empty, use the default operation by raising an error
        if not value_sig:
            raise NotImplementedError

        # If value cannot be iterated over, make it into a list
        if isinstance(value, str) or not hasattr(value, '__iter__'):
            value = [value]

        # Convert value_sig to a list
        value_sig = list(value_sig)

        # Check if the first argument of value_sig is an integer
        if isinstance(value_sig[0], int):
            # If so, use tha tab indicated by this index
            index = value_sig.pop(0)
            tabWidgets = [self.widget(index)]
            indexed = True
        else:
            # If not, use all tabs
            tabWidgets = self.tabWidgets()
            indexed = False

        # Obtain the list of all required boxes
        tabBoxes = [widget.layout().widgets for widget in tabWidgets]

        # Check if the first argument of value_sig is a string
        if value_sig and isinstance(value_sig[0], str):
            # If so, it indicates the key of a specific box
            key = value_sig.pop(0)
            boxes = [tab[key] for tab in tabBoxes]

            # Loop over all boxes and assign the values
            for box, val in zip(boxes, value):
                set_box_value(box, val)

        # If not, it must have been indexed
        elif indexed:
            # If it was indexed, obtain the sole tab that was obtained
            boxes = tabBoxes[0]

            # Loop over all values in value and assign them
            for key, val in value.items():
                set_box_value(boxes[key], val)

        # Else, the signature is invalid, so raise an error
        else:
            raise NotImplementedError


# Define 'MultiData1D' plot property, holding multiple 'Data1D' props
class MultiData1DProp(MultiDataNDProp):
    """
    Provides the definition of the :class:`~MultiData1DProp` plot property.

    This property contains a tab widget with multiple `Data1DProp` properties.

    """

    # Class attributes
    NAME = "MultiData1D"
    REQUIREMENTS = [*MultiDataNDProp.REQUIREMENTS, *Data1DProp.REQUIREMENTS]

    # Initialize multi 1D data property
    def __init__(self, *args, **kwargs):
        # Call super constructor
        super().__init__(Data1DProp, *args, **kwargs)


# Define 'MultiData2D' plot property, holding multiple 'Data2D' props
class MultiData2DProp(MultiDataNDProp):
    """
    Provides the definition of the :class:`~MultiData2DProp` plot property.

    This property contains a tab widget with multiple `Data2DProp` properties.

    """

    # Class attributes
    NAME = "MultiData2D"
    REQUIREMENTS = [*MultiDataNDProp.REQUIREMENTS, *Data2DProp.REQUIREMENTS]

    # Initialize multi 2D data property
    def __init__(self, *args, **kwargs):
        # Call super constructor
        super().__init__(Data2DProp, *args, **kwargs)


# Define 'MultiData3D' plot property, holding multiple 'Data3D' props
class MultiData3DProp(MultiDataNDProp):
    """
    Provides the definition of the :class:`~MultiData3DProp` plot property.

    This property contains a tab widget with multiple `Data3DProp` properties.

    """

    # Class attributes
    NAME = "MultiData3D"
    REQUIREMENTS = [*MultiDataNDProp.REQUIREMENTS, *Data3DProp.REQUIREMENTS]

    # Initialize multi 3D data property
    def __init__(self, *args, **kwargs):
        # Call super constructor
        super().__init__(Data3DProp, *args, **kwargs)


# Create custom class for setting the data column used in plots
class DataColumnBox(GW.DualComboBox):
    # Initialize DataColumnBox class
    def __init__(self, data_table_plugin_obj, parent=None, *args, **kwargs):
        # Save provided data_table_obj
        self.data_table_plugin = data_table_plugin_obj
        self.tab_widget = data_table_plugin_obj.tab_widget

        # Call super constructor
        super().__init__((False, False), r"<html>&rarr;</html>", parent, *args,
                         **kwargs)

    # This function sets up the data column box
    # TODO: Figure out how to stop this from resizing when contents change
    def init(self, *args, **kwargs):
        # Call super setup
        super().init(*args, **kwargs)

        # Extract the two created comboboxes
        tables_box, columns_box = self[:]
        self.tables_box = tables_box
        self.columns_box = columns_box

        # Add items to data tables combobox
        for i, name in enumerate(self.tab_widget.tabNames()):
            tables_box.addItem(name)
            self.set_tables_box_item_tooltip(i, name)

        # Connect signals for tables_box
        self.tab_widget.tabTextChanged.connect(tables_box.setItemText)
        self.tab_widget.tabTextChanged.connect(
            self.set_tables_box_item_tooltip)
        self.tab_widget.tabWasInserted[int, str].connect(tables_box.insertItem)
        self.tab_widget.tabWasInserted[int, str].connect(
            self.set_tables_box_item_tooltip)
        self.tab_widget.tabWasRemoved.connect(tables_box.removeItem)
        get_modified_signal(tables_box, int).connect(
            self.set_columns_box_table)

        # Set initial contents of columns_box
        self.data_table = None
        self.model = None
        self.set_box_value((None, None))

    # This function sets the tooltip of an item in the tables box
    @QC.Slot(int, str)
    def set_tables_box_item_tooltip(self, index, text):
        """
        Sets the tooltip of a specified table with `index` to the given `text`.

        Parameters
        ----------
        index : int
            The logical index of the data table that needs to change its
            tooltip.
        text : str
            The string that needs to be used for this new tooltip.

        """

        self.tables_box.setItemData(index, text, QC.Qt.ToolTipRole)

    # This function sets the data table in the columns box
    @QC.Slot(int)
    def set_columns_box_table(self, index):
        """
        Sets the columns combobox to contain all columns that are contained in
        the data table with `index`.

        Parameters
        ----------
        index : int
            The logical index of the data table whose columns need to be put
            into columns box.

        """

        # Clear the columns_box of all its items
        self.columns_box.clear()

        # If the currently saved data table still exists, disconnect signals
        if(self.data_table_plugin.tab_widget.indexOf(self.data_table) != -1):
            self.model.columnsInserted.disconnect(self.insert_columns)
            self.model.columnsRemoved.disconnect(self.remove_columns)
            self.model.columnNameChanged.disconnect(
                self.columns_box.setItemText)
            self.model.columnNameChanged.disconnect(
                self.set_columns_box_item_tooltip)

        # If currently a data table is selected, obtain its columns
        if(index != -1):
            # Obtain the data table associated with the provided index
            self.data_table = self.data_table_plugin.dataTable(index)
            self.model = self.data_table.model

            # Add all columns in this data table to the columns box
            for i, name in enumerate(self.model.columnNames()):
                self.columns_box.addItem(name)
                self.set_columns_box_item_tooltip(i, name)

            # Connect signals for columns_box
            self.model.columnsInserted.connect(self.insert_columns)
            self.model.columnsRemoved.connect(self.remove_columns)
            self.model.columnNameChanged.connect(self.columns_box.setItemText)
            self.model.columnNameChanged.connect(
                self.set_columns_box_item_tooltip)

        # Else, set data_table and model to None
        else:
            self.data_table = None
            self.model = None

    # This function sets the tooltip of an item in the columns box
    @QC.Slot(int, str)
    def set_columns_box_item_tooltip(self, index, text):
        """
        Sets the tooltip of a specified column with `index` to the given
        `text`.

        Parameters
        ----------
        index : int
            The logical index of the data column that needs to change its
            tooltip.
        text : str
            The string that needs to be used for this new tooltip.

        """

        self.columns_box.setItemData(index, text, QC.Qt.ToolTipRole)

    # This function inserts column display names into the columns box
    @QC.Slot(QC.QModelIndex, int, int)
    def insert_columns(self, parent, first, last):
        """
        Inserts new columns into the columns box between given `first` and
        `last`.

        Parameters
        ----------
        parent : :obj:`~PyQt5.QtCore.QModelIndex` object
            The parent that was used in the data table model.
            This function does not need it.
        first : int
            The logical index of the first column that was inserted.
        last : int
            The logical index of the last column that was inserted.

        """

        # Insert all columns between first and last+1 to the columns box
        for i in range(first, last+1):
            name = self.model.dataColumn(i).name
            self.columns_box.insertItem(i, name)
            self.set_columns_box_item_tooltip(i, name)

    # This function removes column display names from the columns box
    @QC.Slot(QC.QModelIndex, int, int)
    def remove_columns(self, parent, first, last):
        """
        Removes the columns from the columns box between given `first` and
        `last`.

        Parameters
        ----------
        parent : :obj:`~PyQt5.QtCore.QModelIndex` object
            The parent that was used in the data table model.
            This function does not need it.
        first : int
            The logical index of the first column that was removed.
        last : int
            The logical index of the last column that was removed.

        """

        # If the currently set column has been removed, set it to -1
        if get_box_value(self.columns_box, int) in range(first, last+1):
            set_box_value(self.columns_box, -1)

        # Remove all columns between first and last+1 from the columns box
        for i in reversed(range(first, last+1)):
            self.columns_box.removeItem(i)

    # This function retrieves the data table and column currently selected
    def get_box_value(self, *args, **kwargs):
        """
        Returns the currently selected data table and its associated column.

        Returns
        -------
        data_table : :obj:`~guipy.plugins.data_table.widgets.DataTableWidget` \
            object
            The data table that is currently set in this data column box.
        data_column : :obj:`~guipy.plugins.data_table.widgets.DataTableColumn`\
            object
            The data table column in `data_table` that is currently set.

        """

        # Obtain the currently selected column
        column_index = get_box_value(self.columns_box, int)

        # If currently a valid column is selected, return table and column
        if(column_index != -1):
            return(self.data_table, self.model.dataColumn(column_index))
        # Else, return (None, None)
        else:
            return(None, None)

    # This function sets the data table and column
    def set_box_value(self, value):
        """
        Sets the current value of the data table and its associated column to
        `value`.

        Parameters
        ----------
        value : tuple
            A tuple containing the data table and its associated column,
            formatted as `(data_table, data_column)`.

        """

        # If value[0] is None, value is equal to (-1, -1)
        if value[0] is None:
            set_box_value(self.tables_box, -1)
            set_box_value(self.columns_box, -1)
        else:
            table_name = self.tab_widget.tabText(
                self.tab_widget.indexOf(value[0]))
            set_box_value(self.tables_box, table_name)
            set_box_value(self.columns_box, value[1].name)
